import logging

COLORS = {
    'yellow' : '\033[33m',
    'red'    : '\033[31m',
    }
RESET = '\033[0m'

def colored(text, color=None):
    if not color is None:
        text = COLORS[color] + text + RESET
    return text


def setup_logger(name='qondor', fmt=None):
    if name in logging.Logger.manager.loggerDict:
        logger = logging.getLogger(name)
        logger.info('Logger %s is already defined', name)
    else:
        if fmt is None:
            fmt = logging.Formatter(
                fmt = (
                    colored(
                        '[{0}|%(levelname)8s|%(asctime)s|%(module)s]:'.format(name),
                        'yellow'
                        )
                    + ' %(message)s'
                    ),
                datefmt='%Y-%m-%d %H:%M:%S'
                )
        handler = logging.StreamHandler()
        handler.setFormatter(fmt)
        logger = logging.getLogger(name)
        logger.setLevel(logging.DEBUG)
        logger.addHandler(handler)
    return logger


def setup_subprocess_logger():
    return setup_logger(
        'subprocess',
        fmt = logging.Formatter(
            fmt = colored('[%(asctime)s]:', 'red') + ' %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S'
            )
        )


def set_log_file(
        log_file,
        logger_name='qondor',
        subprocess_logger_name='subprocess'
        ):
    """
    Also outputs all logging to a file, but keeps the output
    to stderr as well.
    """
    log_file = osp.abspath(log_file)

    logger = logging.getLogger(logger_name)
    subprocess_logger = logging.getLogger(subprocess_logger_name)

    file_handler = logging.FileHandler(log_file)
    file_handler.setFormatter(LOGGER_FORMATTER)
    logger.addHandler(file_handler)

    # Little bit dangerous; not sure whether logging will open the same file twice
    subprocess_file_handler = logging.FileHandler(log_file)
    subprocess_file_handler.setFormatter(SUBPROCESS_LOGGER_FORMATTER)
    subprocess_logger.addHandler(subprocess_file_handler)




# -*- coding: utf-8 -*-
import qondor, seutils
import logging, re, os.path as osp, datetime, os, math, glob, pprint, json
logger = logging.getLogger('qondor')

def preprocessing(filename):
    """
    Convenience function that returns a Preprocessor object
    """
    return Preprocessor(filename)


def iter_preprocess_lines(lines):
    _linebreak_cache = ''
    for line in lines:
        line = line.strip()
        if line.startswith('#$'):
            line = line.lstrip('#$').strip()
            if len(line) == 0:
                continue
            elif line.endswith('\\'):
                # Line continuation
                _linebreak_cache += line[:-1].strip() + ' '
                logger.debug('Line continuation: set _linebreak_cache to "%s"', _linebreak_cache)
                continue
            elif not _linebreak_cache == '':
                # If there was line continuation and this line does not
                # continue, it must be the last of the continuation
                yield _linebreak_cache + line
                _linebreak_cache = ''
                continue
            yield line

def get_preprocess_lines(lines):
    return list(iter_preprocess_lines(lines))

def iter_preprocess_file(filename):
    with open(filename, 'r') as f:
        # File object should support iteration
        for line in iter_preprocess_lines(f):
            yield line

def get_preprocess_file(filename):
    return list(iter_preprocess_file(filename))


class Preprocessor(object):
    """docstring for Preprocessor"""

    # Cached output from ls calls
    LS_CACHE_FILE = 'ls_cache.json' # Can be overwritten
    LS_CACHE = {}

    @classmethod
    def add_ls_cache(cls, argument, is_se='auto'):
        if argument in cls.LS_CACHE:
            logger.info('Argument "%s" is already cached', argument)
            return cls.get_ls_cache(argument)
        # Check if this concerns a storage element
        if is_se == 'auto': is_se = seutils.has_protocol(argument)
        if is_se:
            contents = seutils.ls_wildcard(argument)
        else:
            contents = [ osp.join(os.getcwd(), i) for i in glob.glob(argument)]
        logger.debug('ls("%s") yielded:\n%s', argument, pprint.pformat(contents))
        cls.LS_CACHE[argument] = contents
        return contents

    @classmethod
    def get_ls_cache(cls, argument):
        return cls.LS_CACHE[argument]

    @classmethod
    def read_ls_cache(cls, cache_file):
        """
        Reads a cache file into LS_CACHE
        """
        logger.info('Reading ls cache file %s', cache_file)
        with open(cache_file, 'r') as f:
            cls.LS_CACHE = json.load(f)

    @classmethod
    def dump_ls_cache(cls, cache_file):
        """
        Dumps the LS_CACHE class dict to a file
        """
        logger.info('Dumping ls cache (%s arguments) to %s', len(cls.LS_CACHE.keys()), cache_file)
        if not qondor.DRYMODE:
            with open(cache_file, 'w') as f:
                json.dump(cls.LS_CACHE, f)

    # Counter for the number of subsets that have been created
    ISET = 0

    allowed_pip_install_instructions = [
        'module-install',
        'pypi-install',
        'install'
        ]

    @classmethod
    def from_lines(cls, lines):
        """For legacy"""
        return cls(lines=lines)

    def __init__(self, filename=None, lines=None, parent=None, ls_cache_file=None):
        super(Preprocessor, self).__init__()
        self.parent = parent
        self.subsets = []
        self.htcondor = {}
        self.pip = []
        self.env = {}
        self.variables = {}
        self.files = {}
        self.items = []
        self.rootfile_chunks = []
        self.delayed_runtime = None
        self.allowed_lateness = None
        # File with cached output from ls processing
        self.ls_cache = {}
        self.ls_cache_file = ls_cache_file
        # Count how many Preprocessor instances there are
        self.subset_index = self.__class__.ISET
        self.__class__.ISET += 1
        if self.is_master():
            # It's the master Preprocessor instance; set some defaults
            self.pip = [
                # Always pip install qondor itself
                ('qondor', self.get_pip_install_instruction('qondor'))
                ]
            if 'el6' in os.uname()[2]:
                logger.info('Detected slc6')
                self.env = {
                    'gccsetup' : '/cvmfs/sft.cern.ch/lcg/contrib/gcc/7/x86_64-slc6-gcc7-opt/setup.sh',
                    'pipdir' : '/cvmfs/sft.cern.ch/lcg/releases/pip/19.0.3-06476/x86_64-slc6-gcc7-opt',
                    'rootsetup' : '/cvmfs/sft.cern.ch/lcg/releases/LCG_95/ROOT/6.16.00/x86_64-slc6-gcc7-opt/ROOT-env.sh',
                    'SCRAM_ARCH' : 'slc6_amd64_gcc700',
                    }
            else:
                self.env = {
                    'gccsetup' : '/cvmfs/sft.cern.ch/lcg/contrib/gcc/7/x86_64-centos7/setup.sh',
                    'pipdir' : '/cvmfs/sft.cern.ch/lcg/releases/pip/19.0.3-06476/x86_64-centos7-gcc7-opt',
                    'rootsetup' : '/cvmfs/sft.cern.ch/lcg/releases/LCG_95/ROOT/6.16.00/x86_64-centos7-gcc7-opt/ROOT-env.sh',
                    'SCRAM_ARCH' : 'slc7_amd64_gcc820',
                    }
        # Read lines if necessary
        if filename and not lines:
            self.filename = osp.abspath(filename)
            self.preprocess()
        elif lines and not filename:
            self.preprocess_lines(lines)
        elif lines and filename:
            raise ValueError('Pass either filename or lines as an init argument')

    def is_master(self):
        return self.parent is None # Should only be true for the master processor

    def merge(self, base):
        """
        Takes base preprocessor variables and adds self on top of it, so that base things are inherited
        """
        # Overwrite style:
        self.htcondor = dict(base.htcondor, **self.htcondor)
        self.variables = dict(base.variables, **self.variables)
        self.env = dict(base.env, **self.env)
        self.files = dict(base.files, **self.files)
        if hasattr(base, 'filename') and not hasattr(self, 'filename'): self.filename = base.filename
        if self.delayed_runtime is None: self.delayed_runtime = base.delayed_runtime
        if self.allowed_lateness is None: self.allowed_lateness = base.allowed_lateness
        # Append style:
        self.items = base.items + self.items
        self.rootfile_chunks = base.rootfile_chunks + self.rootfile_chunks
        self.pip = base.pip + self.pip

    def sets(self):
        """
        Iteratively loops through subsets, updates with parent variables,
        and yields the lowest level subsets (i.e. ones without children).
        Each subset is a full Preprocessor instance.
        """
        if len(self.subsets) == 0:
            yield self
        else:
            for subset in self.subsets:
                subset.merge(self)
                # Iterate
                for s in subset.sets():
                    yield s

    def all_items(self):
        """
        Returns all items of subsets below.
        """
        all_items = []
        for s in self.sets():
            all_items.extend(s.items)
        return all_items

    def all_rootfile_chunks(self):
        """
        Returns all rootfile_chunks of subsets below.
        """
        all_rootfile_chunks = []
        for s in self.sets():
            all_rootfile_chunks.extend(s.rootfile_chunks)
        return all_rootfile_chunks

    def get_pip_install_instruction(self, package_name):
        """
        Checks whether a package is installed editable 
        (i.e. via `pip install -e package`), and if so chooses
        the module-install instruction, which means the editable code will
        be tarred up and sent along with the job. 
        """
        import sys, pkg_resources
        distribution = pkg_resources.get_distribution(package_name)
        if qondor.utils.dist_is_editable(distribution):
            instruction = 'module-install'
        else:
            instruction = 'pypi-install'
        logger.debug('Determined install instruction %s for package %s', instruction, package_name)
        return instruction

    def get_item(self):
        if not(len(self.split_transactions)):
            raise RuntimeError(
                '.get_item() should only be called if transactions are split. '
                'Either .preprocess() is not yet called, or there is no split_transactions '
                'directive.'
                )
        if qondor.BATCHMODE:
            return os.environ['QONDORITEM']
        else:
            logger.debug('Local mode: returning first item of %s', self.split_transactions)
            return self.split_transactions[0]

    def preprocess(self):
        """
        Wrapper for self.preprocess_lines that just takes a path to a file
        rather than direct lines
        """
        # Pass the list rather than the iterator, since sets might make slices
        self.preprocess_lines(get_preprocess_file(self.filename))

    def preprocess_lines(self, lines):
        # Preprocesses a list of lines
        # Starts a subset upon encountering the keyword 'set'
        line_iterator = enumerate(lines)
        for i_line, line in line_iterator:
            logger.debug('Processing l%s: %s', i_line, line)
            line = line.strip()
            if len(line) == 0:
                continue
            elif line.startswith('endset'):
                # This marks the end of a set, so stop preprocessing lines in this processor
                # If this preprocessor is the master, an endset must have been encountered without
                # a matching starting set-line
                if self.is_master():
                    raise ValueError(
                        'Encountered endset without a matching opening set:\n'
                        '--> l{0}: {1}'.format(i_line, line)
                        )
                # Return the line index so that the parent processor can skip the lines processed
                # in this set.
                # Note this line number will be the 'local' line number of the subset of lines
                # that was passed to the child processor
                logger.debug('Closing set at line %s', i_line)
                return i_line
            elif line.startswith('set ') or line == 'set':
                # Start a subset from this line forward
                logger.debug('Starting new set')
                subset = Preprocessor(parent=self)
                self.subsets.append(subset)
                i_line_endset = subset.preprocess_lines(lines[i_line+1:])
                # Skip lines processed in the subset
                for i in range(i_line_endset+1):
                    try:
                        i_line_skip, line_skip = next(line_iterator)
                        logger.debug('Skipping l%s: "%s" (already used in subset)', i_line_skip, line_skip)
                    except StopIteration:
                        # If this happens there is a bug in the code
                        logger.error('StopIteration occured while skipping lines in a subset')
                        raise
                continue
            else:
                self.preprocess_line(line)
        # Only the master should make it to here - subsets should be closed with an endset tag
        if not self.is_master():
            raise ValueError(
                'Encountered EOF without an expected closing endset'
                )

    def preprocess_line(self, line):
        if   line.startswith('htcondor '):
            self.preprocess_line_htcondor(line)
        elif line.startswith('pip '):
            self.preprocess_line_pip(line)
        elif line.startswith('file '):
            self.preprocess_line_file(line)
        elif line.startswith('env '):
            self.preprocess_line_env(line)
        elif line.startswith('delay '):
            self.preprocess_line_delay(line)
        elif line.startswith('allowed_lateness '):
            self.preprocess_line_allowed_lateness(line)
        elif line.startswith('items '):
            self.preprocess_line_items(line)
        else:
            self.preprocess_line_variable(line)

    def preprocess_line_items(self, line):
        """
        Splits a line and returns a list of items, each to be processed in one job.
        If multiple 'items' lines are encountered, items are simply added
        Line elements "b=<number>" and "n=<number>" are considered chunkification parameters
        """
        unprocessed_items = line.split()[1:] # Drop the keyword
        items = []
        n_chunks = None
        chunk_size = None
        chunk_size_events = None
        max_items = None
        for item in unprocessed_items:
            # If an 'item' starts with n= or b=, read that value and don't consider it
            # an item to process. These 'flags' are to configure the chunkification.
            # Only the last parameter counts, any previous parameter will be overwritten.
            # New flag: e= to make chunks with a specific number entries. Only works for
            # for root files.
            # New flag: nmax= to limit the number of items to maximally that number. Usefull
            # mostly for debugging.
            if item.startswith('n='):
                n_chunks = int(item.split('=')[-1])
                chunk_size = None
                chunk_size_events = None
                continue
            elif item.startswith('b='):
                chunk_size = int(item.split('=')[-1])
                n_chunks = None
                chunk_size_events = None
                continue
            elif item.startswith('e='):
                chunk_size_events = int(item.split('=')[-1])
                n_chunks = None
                chunk_size = None
                continue
            elif item.startswith('nmax='):
                max_items = int(item.split('=')[-1])
                continue
            else:
                # It's an item; see if it's meant to be expanded:
                match = re.match(r'ls\((.*)\)', item)
                if match:
                    # Evaluate the ls argument
                    argument = match.group(1)
                    if qondor.BATCHMODE:
                        # Only read the cache, don't execute listing
                        contents = self.get_ls_cache(argument)
                    else:
                        # Add it to the cache
                        contents = self.add_ls_cache(argument)
                    if not len(contents): logger.warning('%s yielded no items', match.group())
                    items.extend(contents)
                else:
                    items.append(item)
        n_items = len(items)
        # Small function to truncate items by nmax if necessary
        def truncate_items(items):
            if max_items and len(items) > max_items:
                logger.info('nmax: Limiting items to %s (from total %s)', max_items, len(items))
                items = items[:max_items]
            return items
        if n_chunks is None and chunk_size is None and chunk_size_events is None:
            # There was no chunkification flag;
            # Use the default of 1 item per chunk, and let's not bother putting it in a list
            self.items.extend(truncate_items(items))
        elif chunk_size_events:
            # The n-entry-sized chunks option was used, make chunks by number of entries
            items = list(seutils.root.iter_chunkify_rootfiles_by_entries(items, chunk_size_events))
            self.rootfile_chunks.extend(truncate_items(items))
        else:
            # Either chunk_size or n_chunks was used
            if chunk_size:
                # If chunk_size is given, calculate the number of chunks that fit in the items
                # Otherwise n_chunks should be given
                n_chunks = int(math.ceil(float(n_items) / chunk_size))
            items = qondor.utils.chunkify(items, n_chunks)
            self.items.extend(truncate_items(items))

    def preprocess_line_htcondor(self, line):
        # Remove 'htcondor' and assume 'key value' structure further on
        try:
            key, value = line.split(None, 2)[1:3]
        except ValueError:
            logger.error('line "%s" did not have expected structure', line)
            raise
        logger.debug('htcondor[%s] = %s', key, value)
        self.htcondor[key] = value

    def preprocess_line_pip(self, line):
        try:
            install_instruction, value = line.split(None, 2)[1:3]
        except ValueError:
            logger.error('line "%s" did not have expected structure', line)
            raise
        if not install_instruction in self.allowed_pip_install_instructions:
            logger.error('pip install_instruction %s is not valid', install_instruction)
            raise ValueError
        # If plain 'install', check whether it's from pypi or an editable package to be tarballed up
        if install_instruction == 'install': install_instruction = self.get_pip_install_instruction(value)
        logger.debug('pip %s %s', install_instruction, value)
        self.pip.append((value, install_instruction))

    def preprocess_line_file(self, line):
        try:
            key, path = line.split(None, 2)[1:3]
        except ValueError:
            logger.error('line "%s" did not have expected structure', line)
            raise
        if seutils.has_protocol(path):
            # Do nothing; files on an SE need no processing
            pass
        elif qondor.BATCHMODE:
            logger.debug('BATCHMODE: %s --> %s', path, osp.basename(path))
            path = osp.basename(path)
        else:
            # Make sure path will be absolute
            # If path is currently relative, assume it's relative w.r.t. to the python file
            if not osp.isabs(path):
                if hasattr(self, 'filename'):
                    # Make sure path is relative to the python file that is preprocessed
                    path = osp.abspath(osp.join(osp.dirname(self.filename), path))
                else:
                    path = osp.abspath(path)
        logger.debug('file[%s] = %s', key, path)
        self.files[key] = path

    def preprocess_line_env(self, line):
        try:
            key, value = line.split(None, 2)[1:3]
        except ValueError:
            logger.error('line "%s" did not have expected structure', line)
            raise
        logger.debug('(environment) %s = %s', key, value)
        self.env[key] = value

    def preprocess_line_variable(self, line):
        try:
            key, value = line.split(None, 1)
        except ValueError:
            logger.error('line "%s" did not have expected structure', line)
            raise
        logger.debug('%s = %s', key, value)
        self.variables[key] = value


    # Special keywords preprocessing

    def preprocess_line_delay_or_lateness(self, line):
        try:
            components = line.split()[1:]
        except ValueError:
            logger.error('line "%s" did not have expected structure', line)
            raise
        unit = 's' if len(components) <= 1 else components[1]
        conversion_to_seconds = {'s' : 1, 'm' : 60, 'h' : 3600}
        if not unit in conversion_to_seconds:
            raise ValueError(
                'Delay unit should be in %s', conversion_to_seconds.keys()
                )
        n_seconds = int(components[0]) * conversion_to_seconds[unit]
        return n_seconds        

    def preprocess_line_delay(self, line):
        n_seconds_delay = self.preprocess_line_delay_or_lateness(line)
        self.delayed_runtime = qondor.utils.get_now_utc() + datetime.timedelta(seconds=n_seconds_delay)
        logger.debug('Jobs will sleep until %s (%s seconds in the future)', self.delayed_runtime, n_seconds_delay)

    def preprocess_line_allowed_lateness(self, line):
        self.allowed_lateness = self.preprocess_line_delay_or_lateness(line)
        logger.debug('Allowed lateness is set to %s seconds', self.allowed_lateness)



# -*- coding: utf-8 -*-
import qondor
import logging, os, os.path as osp, pprint, shutil, uuid, re
from time import strftime
from contextlib import contextmanager
import seutils
logger = logging.getLogger('qondor')


class SHFile(object):
    """
    Writes a .sh file based on a preprocessing object.
    """
    def __init__(self, preprocessing, python_script_args=None):
        self.preprocessing = preprocessing
        self.python_script_args = python_script_args

    def to_file(self, filename, dry=None):
        if dry is None: dry = qondor.DRYMODE
        sh = '\n'.join(self.parse())
        logger.info('Parsed the following .sh file:\n%s', sh)
        logger.info('Writing to %s', filename)
        if not dry:
            with open(filename, 'w') as f:
                f.write(sh)

    def parse(self):
        lines = []
        lines.extend(self.initialize())
        lines.extend(self.set_env_variables())
        lines.extend(self.set_cms_env_and_pip())
        lines.extend(self.pip_installations())
        lines.extend(self.sleep_until_runtime())
        lines.extend(self.python_call())
        return lines

    def initialize(self):
        lines = [
            '#!/bin/bash',
            'set -e',
            'echo "hostname: $(hostname)"',
            'echo "date:     $(date)"',
            'echo "pwd:      $(pwd)"',
            'echo "Redirecting all output to stderr from here on out"',
            'exec 1>&2',
            ''
            ]
        return lines

    def set_env_variables(self):
        lines = []
        for key, value in self.preprocessing.env.items():
            lines.append('export {0}={1}'.format(key, value))
        lines.append('')
        return lines

    def set_cms_env_and_pip(self):
        lines = [
            'export VO_CMS_SW_DIR=/cvmfs/cms.cern.ch/',
            'source /cvmfs/cms.cern.ch/cmsset_default.sh',
            'source "${gccsetup}"',
            'source "${rootsetup}"',
            'export PATH="${pipdir}/bin:${PATH}"',
            'export PYTHONPATH="${pipdir}/lib/python2.7/site-packages/:${PYTHONPATH}"',
            '',
            'set -uxoE pipefail',
            'echo "Setting up custom pip install dir"',
            'HOME="$(pwd)"',
            'pip(){ ${pipdir}/bin/pip "$@"; }  # To avoid any local pip installations',
            'export pip_install_dir="$(pwd)/install"',
            'mkdir -p "${pip_install_dir}/bin"',
            'mkdir -p "${pip_install_dir}/lib/python2.7/site-packages"',
            'export PATH="${pip_install_dir}/bin:${PATH}"',
            'export PYTHONPATH="${pip_install_dir}/lib/python2.7/site-packages:${PYTHONPATH}"',
            '',
            'pip -V',
            'which pip',
            ''
            ]
        return lines

    def pip_installations(self):
        lines = []
        for package, install_instruction in self.preprocessing.pip:
            package = package.replace('.', '-').rstrip('/')
            if install_instruction == 'module-install':
                # Editable install: Manually give tarball, extract, and install
                lines.extend([
                    'mkdir {0}'.format(package),
                    'tar xf {0}.tar -C {0}'.format(package),
                    'pip install --install-option="--prefix=${{pip_install_dir}}" -e {0}/'.format(package)
                    ])
            elif install_instruction == 'pypi-install':
                # Non-editable install from pypi
                lines.append(
                    'pip install --install-option="--prefix=${{pip_install_dir}}" {0}'.format(package)
                    )
        return lines

    def sleep_until_runtime(self):
        if not(self.preprocessing.delayed_runtime is None):
            if not(self.preprocessing.allowed_lateness is None):
                allowed_lateness = '--allowed_lateness {0}'.format(self.preprocessing.allowed_lateness)
            else:
                allowed_lateness = ''
            return [
                'qondor-sleepuntil "{0}" {1}'.format(
                    self.preprocessing.delayed_runtime.strftime('%Y-%m-%d %H:%M:%S'),
                    allowed_lateness
                    ),
                ''
                ]
        else:
            return []

    def python_call(self):
        python_cmd = 'python {0}'.format(osp.basename(self.preprocessing.filename))
        if self.python_script_args:
            # Add any arguments for the python script to this line
            try:  # py3
                from shlex import quote
            except ImportError:  # py2
                from pipes import quote
            python_cmd += ' ' + ' '.join([quote(s) for s in self.python_script_args])
        return [python_cmd, '']


class BaseSubmitter(object):
    """
    Base class for Submitter: Translates user input to one of the following:
    - a submission dict + the number of jobs to submit
    - a submission dict + a list of items
    - a submission dict + a list of chunks
    , and creates a directory with the needed input files for the job
    """

    def __init__(self, python_script_args=None):
        super(BaseSubmitter, self).__init__()
        self.transfer_files = []
        self.python_script_args = python_script_args
        self._created_python_module_tarballs = []

    # _________________________________________________________
    # Methods to prep submission dict

    def format_for_htcondor_interface(self, shfile, preprocessor):
        """
        Formats a submission dict and possibly an extra argument (njobs/items/chunks).
        Returns a tuple (submissiondict, extra_settings), where the extra_settings is
        a dictionary with the following possible keys:
        - 'items' in case there were items found in the preprocessing
        - 'rootfile_chunks' in case there were rootfile_chunks found in the preprocessing
        - 'njobs' in case the njobs variable was set in the preprocessing
        - 'njobsmax'?
        """
        extra_submission_settings = {}
        sub = qondor.schedd.get_default_sub()
        sub['executable'] =  osp.basename(shfile)
        sub['+QondorRundir']  =  '"' + self.rundir + '"'
        sub['environment']['QONDORISET'] = str(preprocessor.subset_index)
        # Overwrite htcondor keys defined in the preprocessing
        sub.update(preprocessor.htcondor)
        # Flatten files into a string, excluding files on storage elements
        transfer_files = self.transfer_files + [f for f in preprocessor.files.values() if not seutils.has_protocol(f)]
        if len(transfer_files) > 0:
            sub['transfer_input_files'] = ','.join(transfer_files)
        # Otherwise continue processing items
        if preprocessor.items and preprocessor.rootfile_chunks:
            raise ValueError(
                'Both regular items and rootfiles-split-by-entries are specified; '
                'This is not supported.'
                )
        elif preprocessor.items:
            logger.info('Items:\n%s', pprint.pformat(preprocessor.items))
            extra_submission_settings['items'] = preprocessor.items
        elif preprocessor.rootfile_chunks:
            logger.info('Items as rootfile chunks:\n%s', pprint.pformat(preprocessor.rootfile_chunks))
            extra_submission_settings['rootfile_chunks'] = preprocessor.rootfile_chunks
        else:
            # Simple case with no items; just use the njobs variable with a default of 1
            if 'njobs' in preprocessor.variables:
                extra_submission_settings['njobs'] = int(preprocessor.variables['njobs'])
        return sub, extra_submission_settings

    # _________________________________________________________
    # Prep functions for files needed for the job

    def tar_python_module(self, module_name):
        """
        Creates a tarball of a python module that is needed inside the job
        """
        if module_name in self._created_python_module_tarballs:
            logger.debug('Not tarring up %s again, tarball already created', module_name)
            return
        logger.info('Creating tarball for python module %s', module_name)
        import importlib
        module = importlib.import_module(module_name)
        tarball = qondor.utils.tarball_python_module(
            module,
            outdir = self.rundir,
            )
        self.transfer_files.append(tarball)
        self._created_python_module_tarballs.append(module_name)

    def make_rundir(self):
        """
        Creates a directory in from which to submit the job
        """
        self.rundir = osp.abspath('qondor_{0}_{1}'.format(
            self.python_name,
            strftime(qondor.TIMESTAMP_FMT)
            ))
        qondor.utils.create_directory(
            self.rundir,
            must_not_exist=True,
            )

    def dump_ls_cache_file(self):
        """
        Creates the ls cache file for the job
        """
        cache_file = osp.join(self.rundir, qondor.Preprocessor.LS_CACHE_FILE)
        qondor.Preprocessor.dump_ls_cache(cache_file)
        self.transfer_files.append(cache_file)

    def dump_rootcache(self):
        """
        Dumps the current state of the seutils root cache to a file to be used in the job
        """
        if seutils.root.USE_CACHE:
            rootcache_file = osp.join(self.rundir, 'rootcache.tar.gz')
            if not qondor.DRYMODE: seutils.root.cache_to_file(rootcache_file)
            self.transfer_files.append(rootcache_file)

    def iter_submissions(self):
        """
        Does all the prep work for the submission directory, and then
        loops over all sets and yields (submission_dict, extra_settings) as
        returned by self.format_for_htcondor_interface
        """
        # First do all prep
        at_least_one_job_submitted = False
        try:
            self.make_rundir()
            self.copy_python_file()
            self.dump_ls_cache_file()
            self.dump_rootcache()
            # Loop over all 'sets'
            # If there are no subsets, this is just a len(1) iterator of the preprocessing
            for i_set, preprocessor in enumerate(self.preprocessing.sets()):
                # Create tarballs for local python modules
                for package, install_instruction in preprocessor.pip:
                    if install_instruction == 'module-install':
                        self.tar_python_module(package)
                # Create the bash script entry point for this job
                shfile = osp.join(self.rundir, '{}_{}.sh'.format(self.python_name, i_set))
                SHFile(preprocessor, self.python_script_args).to_file(shfile)
                # Submit jobs to htcondor
                sub, extra_settings = self.format_for_htcondor_interface(shfile, preprocessor)
                yield sub, extra_settings, preprocessor
                at_least_one_job_submitted = True
        except StopIteration:
            pass
        except Exception as e:
            logger.error('Exception %s was raised', e)
            if not at_least_one_job_submitted:
                logger.error('Error during submission; cleaning up %s', self.rundir)
                if osp.isdir(self.rundir) and not qondor.DRYMODE: shutil.rmtree(self.rundir)
            raise

    def submissions(self):
        return list(iter_submissions)


class Submitter(BaseSubmitter):
    """
    Standard Submitter based on a python file.
    Upon running `.submit()`, will create a new directory,
    transfer all relevant files from the job, and start
    running.
    """
    def __init__(self, python_file, python_script_args=None):
        super(Submitter, self).__init__(python_script_args=python_script_args)
        self.original_python_file = osp.abspath(python_file)
        self.python_base = osp.basename(self.original_python_file)
        self.python_name = self.python_base.replace('.py','')
        self.preprocessing = qondor.Preprocessor(self.original_python_file)

    def copy_python_file(self):
        self.python_file = osp.join(self.rundir, self.python_base)
        qondor.utils.copy_file(self.original_python_file, self.python_file)
        self.transfer_files.append(self.python_file)


class CodeSubmitter(BaseSubmitter):
    """
    Like submitter, but rather than being instantiated from a python file
    it is instantiated from python code in a string
    """
    def __init__(self, python_code, preprocessing_code=None, name='', dry=False):
        super(CodeSubmitter, self).__init__(dry)
        self.python_code = python_code.split('\n') if qondor.utils.is_string(python_code) else python_code
        if preprocessing_code is None: preprocessing_code = []
        self.preprocessing_code = preprocessing_code.split('\n') if qondor.utils.is_string(preprocessing_code) else preprocessing_code
        self.python_name = name if name else 'fromcode'
        self.preprocessing = qondor.Preprocessor.from_lines(self.preprocessing_code)

    def copy_python_file(self):
        self.python_file = osp.join(self.rundir, 'pythoncode.py')
        logger.info('Writing %s lines of code to %s', len(self.python_code), self.python_file)
        with open(self.python_file, 'w') as f:
            f.write('\n'.join(['#$ ' + l for l in self.preprocessing_code]))
            f.write('\n')
            f.write('\n'.join(self.python_code))
        self.transfer_files.append(self.python_file)
        self.preprocessing.filename = self.python_file

    def run_local_exec(self):
        logger.warning('Doing exec() on the following code:\n%s', '\n'.join(self.python_code))
        if '__file__' in self.python_code:
            logger.error(
                'Note that "__file__" will crash, as it is not defined in exec(). '
                'Use .run_local() instead.'
                )
        logger.warning('Output:')
        exec('\n'.join(self.python_code))

    def run_local(self):
        python_file = 'temp-{0}.py'.format(uuid.uuid4())
        logger.info('Put python code in {0} and running:'.format(python_file))
        try:
            with open(python_file, 'w') as f:
                f.write('\n'.join(['#$ ' + l for l in self.preprocessing_code]))
                f.write('\n')
                f.write('\n'.join(self.python_code))
            return qondor.utils.run_command(['python', python_file])
        finally:
            logger.info('Removing %s', python_file)
            os.remove(python_file)


def cmsconnect_settings(sub, preprocessor=None, cli=False):
    """
    Adds special cmsconnect settings to submission dict in order to submit
    on cmsconnect via the python bindings, or in case of submitting by the
    command line, to set the DESIRED_Sites key.
    Modifies the dict in place.
    """
    # Read the central config for cmsconnect
    try:
        from configparser import RawConfigParser # python 3
    except ImportError:
        import ConfigParser # python 2
        RawConfigParser = ConfigParser.RawConfigParser
    cfg = RawConfigParser()
    cfg.read('/etc/ciconnect/config.ini')
    sites = cfg.get('submit', 'DefaultSites').split(',')
    all_sites = set(sites)
    # Check whether the user whitelisted or blacklisted some sites
    if preprocessor:
        import fnmatch
        blacklisted = []
        whitelisted = []
        # Build the blacklist
        if 'sites_blacklist' in preprocessor.variables:
            for blacksite in preprocessor.variables['sites_blacklist'].split():
                for site in sites:
                    if fnmatch.fnmatch(site, blacksite):
                        blacklisted.append(site)
        # Build the whitelist
        if 'sites_whitelist' in preprocessor.variables:
            for whitesite in preprocessor.variables['sites_whitelist'].split():
                for site in sites:
                    if fnmatch.fnmatch(site, whitesite):
                        whitelisted.append(site)
        blacklisted = list(set(blacklisted))
        blacklisted.sort()
        whitelisted = list(set(whitelisted))
        whitelisted.sort()
        logger.info('Blacklisting: %s', ','.join(blacklisted))
        logger.info('Whitelisting: %s', ','.join(whitelisted))
        sites = list( (set(sites) - set(blacklisted)).union(set(whitelisted)) )
        sites.sort()
    logger.info('Submitting to sites: %s', ','.join(sites))
    # Add a plus only if submitting via .jdl file
    addplus = lambda key: '+' + key if cli else key
    if all_sites != set(sites):
        sub[addplus('DESIRED_Sites')] = '"' + ','.join(sites) + '"'
    if not cli:
        sub[addplus('ConnectWrapper')] = '"2.0"'
        sub[addplus('CMSGroups')] = '"/cms,T3_US_FNALLPC"'
        sub[addplus('MaxWallTimeMins')] = '500'
        sub[addplus('ProjectName')] = '"cms.org.fnal"'
        sub[addplus('SubmitFile')] = '"irrelevant.jdl"'
        sub[addplus('AccountingGroup')] = '"analysis.{0}"'.format(os.environ['USER'])
        logger.warning('FIXME: CMS Connect settings currently hard-coded for a FNAL user')



from setuptools import setup

setup(
    name          = 'qondor',
    version       = '0.63',
    license       = 'BSD 3-Clause License',
    description   = 'Description text',
    url           = 'https://github.com/tklijnsma/qondor.git',
    download_url  = 'https://github.com/tklijnsma/qondor/archive/v0_63.tar.gz',
    author        = 'Thomas Klijnsma',
    author_email  = 'tklijnsm@gmail.com',
    packages      = ['qondor'],
    zip_safe      = False,
    scripts       = [
        'bin/qondor-submit',
        'bin/qondor-sleepuntil'
        ],
    install_requires     = ['seutils'],
    include_package_data = True,
    )



# -*- coding: utf-8 -*-
import os, os.path as osp, uuid
from .logger import setup_logger, setup_subprocess_logger
logger = setup_logger()
subprocess_logger = setup_subprocess_logger()

INCLUDE_DIR = osp.join(osp.abspath(osp.dirname(__file__)), 'include')

from . import utils

BATCHMODE = False
if 'QONDOR_BATCHMODE' in os.environ:
    BATCHMODE = True

COLLECTOR_NODES = None
DEFAULT_MGM = None
TIMESTAMP_FMT = '%Y%m%d_%H%M%S'

from . import schedd
from .schedd import get_best_schedd, get_schedd_ads, wait, remove_jobs
from .preprocess import Preprocessor, preprocessing
from .submit import SHFile, Submitter, CodeSubmitter
from .cmssw import CMSSW
from .cmssw_releases import get_arch
import seutils


# ___________________________________________________
# Decisions at import-time depending on whether this is a job or
# an import on an interactive node

if BATCHMODE:
    # Sets default amount of times to retry a seutils.cp statement
    seutils.N_COPY_RETRIES = 2

# Read cached ls calls if this is a job
if BATCHMODE and osp.isfile(Preprocessor.LS_CACHE_FILE):
    Preprocessor.read_ls_cache(Preprocessor.LS_CACHE_FILE)

# Read rootcache is this is a job
if BATCHMODE and osp.isfile('rootcache.tar.gz'):
    seutils.root.load_cache('rootcache.tar.gz')

# FNAL-specific things
if os.environ.get('HOSTNAME', '').endswith('fnal.gov'):
    # Fix to be able to import htcondor python bindings
    logger.warning('Detected FNAL: Modifying path to use system htcondor python bindings')
    import sys
    if sys.version_info.major < 3:
        sys.path.extend([
            '/usr/lib64/python2.6/site-packages',
            '/usr/lib64/python2.7/site-packages'
            ])
    schedd.GLOBAL_SCHEDDMAN_CLS = schedd.ScheddManagerFermiHTC
    seutils.set_default_mgm('root://cmseos.fnal.gov')


# ___________________________________________________
# 'Globals'

DRYMODE = False
def drymode(flag=True):
    global DRYMODE
    DRYMODE = flag


# ___________________________________________________
# Short cuts

def get_proc_id():
    if BATCHMODE:
        return int(os.environ['CONDOR_PROCESS_ID'])
    else:
        logger.info('Local mode - return proc_id 0')
        return 0

def get_cluster_id():
    if BATCHMODE:
        return os.environ['CONDOR_CLUSTER_NUMBER']
    else:
        logger.info('Local mode - return cluster_id 1234')
        return str(1234)

CACHED_PREPROC = None

def get_master_preproc(python_file=None):
    """
    Returns and caches the preprocessing of a python file.
    By default it picks the top level python file in the stack, and calls the preprocessor on that.
    """
    global CACHED_PREPROC
    if CACHED_PREPROC: return CACHED_PREPROC
    if not python_file:
        import traceback
        stack = traceback.extract_stack()
        python_file = stack[0][0]
    CACHED_PREPROC = preprocessing(python_file)
    return CACHED_PREPROC

def get_preproc():
    """
    Gets the preprocessing set that this job is supposed to do.
    In local mode it just returns the first set.
    If no sets are defined, it returns just the preprocessing.
    """
    if BATCHMODE:
        iset = int(os.environ['QONDORISET'])
        for subset in get_master_preproc().sets():
            if subset.subset_index == iset:
                return subset
        else:
            raise RuntimeError(
                'Looking for set {0} failed: no such set'
                .format(iset)
                )
    else:
        logger.info('Local mode - returning simply first set')
        return next(get_master_preproc().sets())

def get_item():
    """
    Gets the item (string or list of strings) that this job is supposed to do.
    In local mode it just returns the first item
    """
    if BATCHMODE:
        try:
            item = os.environ['QONDORITEM']
            if item.startswith(','):
                # This is a hack to make sure that if the chunk_size is set to 1,
                # get_item() still returns a len-1 list, and not a plain string.
                # Only items meant to be len-1 lists should start with a comma.
                return [item[1:]]
            elif ',' in item:
                # Normal case for a list of items (n_items > 1)
                return item.split(',')
            else:
                # Normal case for a simple string being an item
                return item
        except KeyError:
            logger.error('No item was found for this job! Are you sure you passed items?')
            raise
    else:
        logger.info('Local mode - returning first item')
        items = get_master_preproc().all_items()
        if not items: raise RuntimeError('No items determined in preprocessing')
        return items[get_proc_id()]

def get_chunk():
    """
    Gets the chunk of rootfiles with first and last entries to use (summing up
    to a specified number of entries).
    Format is [ (rootfile, first, last), (rootfile, first, last) ]
    """
    if BATCHMODE:
        try:
            chunk_str = os.environ['QONDORROOTFILECHUNK']
            def read_chunk_str(chunk_str):
                for part in chunk_str.split(';'):
                    rootfile, first, last, is_whole_file = part.split(',')
                    yield rootfile, int(first), int(last), bool(int(is_whole_file)) # Note bool('0') would be True
            return list(read_chunk_str(chunk_str))
        except KeyError:
            logger.error(
                'No chunk was found for this job! Are you sure you passed "items e=..."'
                ', and not "items ..."?'
                )
            raise
    else:
        logger.info('Local mode - returning first chunk')
        chunks = get_master_preproc().all_rootfile_chunks()
        if not chunks: raise RuntimeError('No chunks determined in preprocessing')
        chunk = chunks[get_proc_id()]
        logger.info('Chunk: %s', chunk)
        return chunk

def get_chunk_as_rootfile(dst='chunk.root', cmssw=None):
    """
    Instead of returning the chunk as a list, it takes the chunk and
    makes one single root file out of it by splitting and hadding the
    parts of the chunk.
    If `cmssw` is passed, the splitting is performed in the CMSSW
    environment (which is usually needed for CMSSW root files).
    """
    chunk = get_chunk()
    if cmssw:
        seutils.root.hadd_chunk_entries(chunk, dst, file_split_fn=cmssw.make_chunk_rootfile)
    else:
        seutils.root.hadd_chunk_entries(chunk, dst)
    return dst

def init_cmssw(tarball_key='cmssw_tarball', scram_arch=None):
    """
    A shortcut function to quickly extract and setup a CMSSW tarball
    """
    cmssw_tarball = get_preproc().files[tarball_key]
    cmssw = CMSSW.from_tarball(cmssw_tarball, scram_arch)
    return cmssw

def get_var(variable):
    """
    Shortcut to get a variable defined in the preprocessing
    """
    return get_preproc().variables[variable]

def get_submission_timestamp():
    """
    Returns the submission time as a timestamp string with format TIMESTAMP_FMT
    """
    if BATCHMODE:
        return os.environ['CLUSTER_SUBMISSION_TIMESTAMP']
    else:
        from datetime import datetime
        logger.info('Local mode - returning current time')
        return datetime.now().strftime(TIMESTAMP_FMT)

def get_submission_time():
    """
    Returns the submission time as datetime object
    """
    from datetime import datetime
    if BATCHMODE:
        return datetime.strptime(get_submission_timestamp(), TIMESTAMP_FMT)
    else:
        logger.info('Local mode - returning current time')
        return datetime.now()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import qondor
import logging, re, pprint, os, os.path as osp
from time import sleep, strftime
from contextlib import contextmanager
logger = logging.getLogger('qondor')


CACHED_FUNCTIONS = []
def cache_return_value(func):
    """
    Decorator that only calls a function once, and
    subsequent calls just return the cached return value
    """
    global CACHED_FUNCTIONS
    def wrapper(*args, **kwargs):
        if not getattr(wrapper, 'is_called', False):
            wrapper.is_called = True
            wrapper.cached_return_value = func(*args, **kwargs)
            CACHED_FUNCTIONS.append(wrapper)
        else:
            logger.debug(
                'Returning cached value for %s: %s',
                func.__name__, wrapper.cached_return_value
                )
        return wrapper.cached_return_value
    return wrapper

def clear_cache():
    global CACHED_FUNCTIONS
    for func in CACHED_FUNCTIONS:
        func.is_called = False
    CACHED_FUNCTIONS = []


class ScheddManager(object):
    """
    Class that interacts with htcondor scheduler daemons ('schedds').
    Will first run `get_collector_node_addresses` to obtain htcondor collectors,
    and will get the schedd_ads from the collectors.
    The algorithm for `get_best_schedd` is based on the weight calculation
    implemented on the Fermilab HTCondor cluster.

    Most methods are decorated with `cache_return_value`, meaning subsequent calls
    will return the same value as the first call until the cache is cleared.
    """

    def __init__(self):
        super(ScheddManager, self).__init__()
        self.schedd_ads = []
        self.schedd_constraints = None

    def clear_cache(self):
        """Convenience function to not have to import clear_cache everywhere"""
        clear_cache()

    @cache_return_value
    def get_collector_node_addresses(self):
        """
        Sets the attribute self.collector_node_addresses to a list of addresses to
        htcondor collectors. The collectors subsequently are able to return
        schedulers to which one can submit.

        First looks if qondor.COLLECTOR_NODES is set to anything. If so, it takes the
        addresses from that, allowing the user to easily configure custom collector nodes.

        Otherwise, it tries to obtain addresses from the htcondor parameter COLLECTOR_HOST_STRING.
        """
        import htcondor

        if not(qondor.COLLECTOR_NODES is None):
            logger.info('Setting collectors to qondor.COLLECTOR_NODES %s', qondor.COLLECTOR_NODES)
            self.collector_node_addresses = qondor.COLLECTOR_NODES
            return

        err_msg = (
            'Could not find any collector nodes. '
            'Either set qondor.COLLECTOR_NODES manually to a \', \'-separated '
            'list of addresses, or set the htcondor parameter \'COLLECTOR_HOST_STRING\','
            ' or subclass qondor.schedd.ScheddManager so that it can find the collector nodes.'
            )
        try:
            collector_node_addresses = htcondor.param['COLLECTOR_HOST_STRING'].strip().strip('"')
        except:
            logger.error(err_msg)
            raise
        if collector_node_addresses is None:
            RuntimeError(err_msg)

        logger.debug('Found collector nodes %s', collector_node_addresses)
        # Seems like a very obfuscated way of writing ".split()" but keep it for now
        self.collector_node_addresses = re.findall(r'[\w\/\:\/\-\/\.]+', collector_node_addresses)
        logger.debug('Using collector_node_addresses %s', self.collector_node_addresses)

    @cache_return_value
    def get_schedd_ads(self):
        import htcondor
        schedd_ad_projection = [
            'Name', 'MyAddress', 'MaxJobsRunning', 'ShadowsRunning',
            'RecentDaemonCoreDutyCycle', 'TotalIdleJobs'
            ]
        try:
            logger.debug('First trying default collector and schedd')
            # This is most likely to work for most batch systems
            collector = htcondor.Collector()
            limited_schedd_ad = collector.locate(htcondor.DaemonTypes.Schedd)
            logger.debug('Retrieved limited schedd ad:\n%s', pprint.pformat(limited_schedd_ad))
            self.schedd_ads = collector.query(
                    htcondor.AdTypes.Schedd,
                    projection = schedd_ad_projection,
                    constraint = 'MyAddress=?="{0}"'.format(limited_schedd_ad['MyAddress'])
                    )
        except Exception as e:
            logger.debug('Default collector and schedd did not work:\n%s\nTrying via collector host string', e)
            self.get_collector_node_addresses()
            for node in self.collector_node_addresses:
                logger.debug('Querying %s for htcondor.AdTypes.Schedd', node)
                collector = htcondor.Collector(node)
                try:
                    self.schedd_ads = collector.query(
                        htcondor.AdTypes.Schedd,
                        projection = schedd_ad_projection,
                        constraint = self.schedd_constraints
                        )
                    if self.schedd_ads:
                        # As soon as schedd_ads are found in one collector node, use those
                        # This may not be the correct choice for some batch systems
                        break
                except Exception as e:
                    logger.debug('Failed querying %s: %s', node, e)
                    continue
            else:
                logger.error('Failed to collect any schedds from %s', self.collector_node_addresses)
                raise RuntimeError

        logger.debug('Found schedd ads: \n%s', pprint.pformat([dict(d) for d in self.schedd_ads]))
        return self.schedd_ads

    @cache_return_value
    def get_best_schedd(self):
        import htcondor
        self.get_schedd_ads()
        if len(self.schedd_ads) == 1:
            best_schedd_ad = self.schedd_ads[0]
        else:
            self.schedd_ads.sort(key = self.get_schedd_weight)
            best_schedd_ad = self.schedd_ads[0]
        logger.debug('Best schedd is %s', best_schedd_ad['Name'])
        schedd = htcondor.Schedd(best_schedd_ad)
        return schedd

    @cache_return_value
    def get_all_schedds(self):
        import htcondor
        return [ htcondor.Schedd(schedd_ad) for schedd_ad in self.get_schedd_ads() ]

    def get_schedd_weight(self, schedd_ad):
        duty_cycle = schedd_ad['RecentDaemonCoreDutyCycle'] * 100
        occupancy = (schedd_ad['ShadowsRunning'] / schedd_ad['MaxJobsRunning']) * 100
        n_idle_jobs = schedd_ad['TotalIdleJobs']
        weight = 0.7 * duty_cycle + 0.2 * occupancy + 0.1 * n_idle_jobs
        logger.debug(
            'Weight calc for %s: weight = %s, '
            'duty_cycle = %s, occupancy = %s, n_idle_jobs = %s',
            schedd_ad['Name'], weight, duty_cycle, occupancy, n_idle_jobs
            )
        return weight


class ScheddManagerFermiHTC(ScheddManager):
    """
    Subclass of ScheddManager specifically for the Fermilab HTCondor setup
    """
    def __init__(self):
        super(ScheddManagerFermiHTC, self).__init__()
        self.schedd_constraints = 'FERMIHTC_DRAIN_LPCSCHEDD=?=FALSE && FERMIHTC_SCHEDD_TYPE=?="CMSLPC"'

    @cache_return_value
    def get_collector_node_addresses(self):
        import htcondor
        try:
            collector_node_addresses = htcondor.param['FERMIHTC_REMOTE_POOL']
            self.collector_node_addresses = re.findall(r'[\w\/\:\/\-\/\.]+', collector_node_addresses)
            logger.info('Set collector_node_addresses to %s', collector_node_addresses)
        except KeyError:
            super(ScheddManagerFermiHTC, self).get_collector_node_addresses()


# -------------- Convenience functions
# Define a GLOBAL_SCHEDDMAN, so that for basic usage one
# can simply call e.g. 'get_best_schedd'

GLOBAL_SCHEDDMAN_CLS = ScheddManager
GLOBAL_SCHEDDMAN = None

def _get_scheddman(renew):
    global GLOBAL_SCHEDDMAN
    if GLOBAL_SCHEDDMAN is None:
        # Create a new instance
        GLOBAL_SCHEDDMAN = GLOBAL_SCHEDDMAN_CLS()
    if renew:
        GLOBAL_SCHEDDMAN.clear_cache()
    return GLOBAL_SCHEDDMAN

def get_best_schedd(renew=False):
    return _get_scheddman(renew).get_best_schedd()

def get_schedd_ads(renew=False):
    return _get_scheddman(renew).get_schedd_ads()

def get_schedds(renew=False):
    return _get_scheddman(renew).get_all_schedds()


# Some basic condor utilities

def get_jobs(cluster_id, proc_id=None):
    requirements = 'ClusterId=={0}'.format(cluster_id)
    if not(proc_id is None): requirements += ' && ProcId == {0}'.format(proc_id)
    classads = []
    logger.debug('requirements = %s', requirements)
    for schedd in get_schedds():
        classads.extend(list(schedd.xquery(
            requirements = requirements,
            projection = ['ProcId', 'JobStatus']
            )))
    logger.info('Query returned %s', classads)
    return classads

def remove_jobs(cluster_id):
    import htcondor
    logger.info('Removing cluster_id %s', cluster_id)
    for schedd in get_schedds():
        schedd.act(htcondor.JobAction.Remove, 'ClusterId=?="{0}"'.format(cluster_id))

def wait(cluster_id, proc_id=None, n_sleep=10):
    while True:
        states = [j['JobStatus'] for j in get_jobs(cluster_id, proc_id)]
        is_done = all([ not(state in [1, 2]) for state in states ])
        if is_done:
            logger.info('ClusterId %s ProcId %s seems done', cluster_id, proc_id)
            break
        else:
            logger.info(
                'ClusterId %s ProcId %s not yet done:\n%s',
                cluster_id, proc_id, states
                )
            logger.info('Sleeping for %s seconds before checking again', n_sleep)
            sleep(n_sleep)

# _____________________________________________________________________
# Submission utils

def get_default_sub():
    """
    Returns the default submission dict (the equivalent of a .jdl file)
    to be used by the submitter. Implemented like this to be subclassable.
    """
    sub = {
        'universe' : 'vanilla',
        'output' : 'out_$(Cluster)_$(Process).txt',
        'error' : 'err_$(Cluster)_$(Process).txt',
        'log' : 'log_$(Cluster)_$(Process).txt',
        'should_transfer_files' : 'YES',
        'environment' : {
            'QONDOR_BATCHMODE' : '1',
            'CONDOR_CLUSTER_NUMBER' : '$(Cluster)',
            'CONDOR_PROCESS_ID' : '$(Process)',
            'CLUSTER_SUBMISSION_TIMESTAMP' : strftime(qondor.TIMESTAMP_FMT),
            },
        }
    # Try to set some more things
    try:
        sub['x509userproxy'] = os.environ['X509_USER_PROXY']
    except KeyError:
        try:
            sub['x509userproxy'] = qondor.utils.run_command(['voms-proxy-info', '-path'])[0].strip()
            logger.info('Set x509userproxy to "%s" based on output from voms-proxy-info', sub['x509userproxy'])
        except:
            logger.warning(
                'Could not find a x509userproxy to pass; manually '
                'set the htcondor variable \'x509userproxy\' if your '
                'htcondor setup requires it.'
                )
    try:
        sub['environment']['USER'] = os.environ['USER']
    except KeyError:
        # No user specified, no big deal
        pass
    return sub

def _change_submitobject_env_variable(submitobject, key, value):
    """
    Tries to replace an environment variable in a Submit-object.
    This hack is needed to have items be in the same cluster.
    """
    env = submitobject['environment']
    new_env = re.sub(key + r'=\'.*?\'', '{0}=\'{1}\''.format(key, value), env)
    logger.debug('Replacing:\n  %s\n  by\n  %s', env, new_env)
    submitobject['environment'] = new_env

def _htcondor_format_environment(env):
    """
    Takes a dict of key : value pairs that are both strings, and
    returns a string that is formatted so that htcondor can turn it
    into environment variables
    """
    return ('"' +
        ' '.join(
            [ '{0}=\'{1}\''.format(key, value) for key, value in env.items() ]
            )
        + '"'
        )

@contextmanager
def _transaction(schedd, dry=None):
    """
    Wrapper for schedd.transaction, + the ability for dry mode
    """
    if dry is None: dry = qondor.DRYMODE
    try:
        if not dry:
            with schedd.transaction() as transaction:
                yield transaction
        else:
            yield None
    finally:
        pass

def _format_item(item):
    """
    Format an item (most likely a list) to a flat string (which will read by the job)
    """
    # Ensure the string will be ',' separated (a len-1 list will not have a comma)
    # Also ensure there are no quotes in it, which would screw up condor's reading
    if qondor.utils.is_string(item):
        item = [item]
    elif len(item) == 1:
        # Exceptional case: a len-1 list passed is indistinguishable from just a string
        # If chunk_size is set to 1, it is intended that QONDORITEM is a len-1 list,
        # but the ','.join() statement makes the item look like a simple string.
        # Add an initial comma as a hack to tell qondor that the item is meant to be a
        # len-1 list.
        item[0] = ',' + item[0]
    item = ','.join([i.replace('\'','').replace('"','') for i in item])
    return item

def _format_chunk(chunk):
    """
    Format a rootfile chunk (most likely a list) to a flat string (which will read by the job)
    """
    # Set the chunk as a string in the environment as follows:
    # rootfile,first,last,is_whole_file;rootfile,first,last,is_whole_file;...
    # Convert is_whole_file to either 0 or 1 first (converting bool to str
    # yields 'True' or 'False', but str('False') yields True...)
    return ';'.join(
        [','.join([ str(e[0]), str(e[1]), str(e[2]), str(int(e[3])) ]) for e in chunk]
        )

def submit_pythonbindings(
    submissiondict, submissiondir='.', schedd=None, dry=None,
    items=None, rootfile_chunks=None, njobs=1, njobsmax=None
    ):
    """
    Main interface to htcondor via the python bindings.
    """
    if dry is None: dry = qondor.DRYMODE # Inherit global drymode flag
    import htcondor
    if schedd is None: schedd = get_best_schedd(renew=True)
    sub = submissiondict.copy() # Create a copy to keep original dict unmodified
    # List to store all submitted job ads
    job_ads = []
    cluster_id = [0]
    n_queued = [0] # Make it a 1-element list so that it's a 'pointer'
    with qondor.utils.switchdir(submissiondir):
        # Make the transaction
        with _transaction(schedd) as transaction:
            # Helper function to actually submit a job to this transaction
            # Modifies the 'global' job_ads and cluster_id variables
            def queue(submit_object, njobs=1):
                ads = []
                n_queued[0] += njobs
                if dry: return
                cluster_id[0] = int(submit_object.queue(transaction, njobs, ads))
                job_ads.extend(ads)
            # Helper function to test whether the needed amount of jobs to submit
            # is already reached
            def should_quit_now():
                if not(njobsmax is None) and n_queued[0] >= njobsmax:
                    return True
                return False
            # Choose specific submit behavior based on what extra keywords were passed
            if items:
                # Items logic: Turn any potential list into a ','-separated string,
                # and set the environment variable QONDORITEM to that string.
                sub['environment']['QONDORITEM'] = 'placeholder' # Placeholder
                sub['environment'] = _htcondor_format_environment(sub['environment'])
                submit_object = htcondor.Submit(sub)
                for item in items:
                    # Change the env variable in the submitobject
                    _change_submitobject_env_variable(submit_object, 'QONDORITEM', _format_item(item))
                    # Submit again and record cluster_id and job ads
                    queue(submit_object)
                    if should_quit_now(): break
            elif rootfile_chunks:
                sub['environment']['QONDORROOTFILECHUNK'] = 'placeholder'
                sub['environment'] = _htcondor_format_environment(sub['environment'])
                submit_object = htcondor.Submit(sub)                
                for chunk in rootfile_chunks:
                    _change_submitobject_env_variable(submit_object, 'QONDORROOTFILECHUNK', _format_chunk(chunk))
                    queue(submit_object)
                    if should_quit_now(): break
            else:
                sub['environment'] = _htcondor_format_environment(sub['environment'])
                submit_object = htcondor.Submit(sub)
                queue(submit_object, njobs if (njobsmax is None) else min(njobs, njobsmax))
    logger.info('Submitted %s jobs to cluster %s', n_queued[0], cluster_id)
    return cluster_id[0], n_queued[0] if dry else job_ads

def submit_condor_submit_commandline(
    submissiondict, submissiondir='.', dry=None,
    items=None, rootfile_chunks=None, njobs=1, njobsmax=None,
    do_not_submit=False
    ):
    """
    Main interface to htcondor via the condor_submit command line tool.
    """
    if dry is None: dry = qondor.DRYMODE # Inherit global drymode flag
    sub = submissiondict.copy() # Create a copy to keep original dict unmodified
    # List to store all submitted job ads
    job_ads = []
    cluster_id = 0
    n_queued = 0
    # Helper function to test whether we should stop submitting now
    def should_quit_now():
        return not(njobsmax is None) and n_queued >= njobsmax
    with qondor.utils.switchdir(submissiondir):
        # Write the .jdl file
        jdl_file = osp.splitext(submissiondict['executable'])[0] + '.jdl'
        logger.debug('Writing submission to %s', jdl_file)
        with qondor.utils.openfile(jdl_file, 'w') as jdl:
            # Write (most) keys in the submission dict to a file
            for key in submissiondict.keys():
                if key.lower() == 'environment': continue
                val = submissiondict[key]
                jdl.write('{} = {}\n'.format(key, val))
            # Choose specific submit behavior based on what extra keywords were passed
            if items:
                # Items logic: Turn any potential list into a ','-separated string,
                # and set the environment variable QONDORITEM to that string.
                for item in items:
                    sub['environment']['QONDORITEM'] = _format_item(item)
                    jdl.write('\nenvironment = {}\n'.format(_htcondor_format_environment(sub['environment'])))
                    jdl.write('queue\n')
                    n_queued += 1
                    if should_quit_now(): break
            elif rootfile_chunks:
                for chunk in rootfile_chunks:
                    sub['environment']['QONDORROOTFILECHUNK'] = _format_chunk(chunk)
                    jdl.write('\nenvironment = {}\n'.format(_htcondor_format_environment(sub['environment'])))
                    jdl.write('queue\n')
                    n_queued += 1
                    if should_quit_now(): break
            else:
                if not(njobsmax is None): njobs = min(njobsmax, njobs)
                jdl.write('environment = {}\n'.format(_htcondor_format_environment(sub['environment'])))
                jdl.write('queue {}\n'.format(njobs))
                n_queued += njobs
        logger.debug('Queued %s jobs', n_queued)
        # Quit now if we don't actually want to submit
        if do_not_submit:
            logger.debug('do_not_submit == True, quiting now')
            return 0, n_queued
        # Run condor_submit on the jdl file (still in the submission directory)
        output = qondor.utils.run_command(['condor_submit', jdl_file], env=qondor.utils.get_clean_env())
        if dry:
            logger.info('Submitted %s jobs to cluster_id 0', n_queued)
            return 0, n_queued
        match = re.search(r'(\d+) job\(s\) submitted to cluster (\d+)', '\n'.join(output))
        if match:
            n_submitted = int(match.group(1))
            cluster_id = match.group(2)
            logger.info('Submitted %s jobs to cluster_id %s', n_submitted, cluster_id)
            return cluster_id, n_submitted
        else:
            logger.error(
                'condor_submit exited ok but could not determine where and how many jobs were submitted'
                )
            return 0, n_queued



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os, shutil, logging, subprocess, glob, pprint, time, datetime, sys
import os.path as osp
from contextlib import contextmanager
import qondor
logger = logging.getLogger('qondor')
subprocess_logger = logging.getLogger('subprocess')

# Python 2.6 compatibiltiy (see https://stackoverflow.com/a/13160748/9209944)
if "check_output" not in dir( subprocess ): # duck punch it in!
    logger.warning('Duck punching subprocess.check_output; suboptimal!')
    def f(*popenargs, **kwargs):
        if 'stdout' in kwargs:
            raise ValueError('stdout argument not allowed, it will be overridden.')
        process = subprocess.Popen(stdout=subprocess.PIPE, *popenargs, **kwargs)
        output, unused_err = process.communicate()
        retcode = process.poll()
        if retcode:
            cmd = kwargs.get("args")
            if cmd is None:
                cmd = popenargs[0]
            raise subprocess.CalledProcessError(retcode, cmd)
        return output
    subprocess.check_output = f

def _create_directory_no_checks(dirname, dry=None):
    """
    Creates a directory without doing any further checks.

    :param dirname: Name of the directory to be created
    :type dirname: str
    :param dry: Don't actually create the directory, only log
    :type dry: bool, optional
    """
    if dry is None: dry = qondor.DRYMODE
    logger.warning('Creating directory %s', dirname)
    if not dry: os.makedirs(dirname)

def create_directory(dirname, renew=False, must_not_exist=False, dry=None):
    """
    Creates a directory if certain conditions are met.

    :param dirname: Name of the directory to be created
    :type dirname: str
    :param renew: Removes the directory `dirname` if it already exists
    :type renew: bool, optional
    :param must_not_exist: Throw an OSError if the directory already exists
    :type must_not_exist: bool, optional
    :param dry: Don't actually create the directory, only log
    :type dry: bool, optional
    """
    if dry is None: dry = qondor.DRYMODE
    if osp.isfile(dirname):
        raise OSError('{0} is a file'.format(dirname))
    isdir = osp.isdir(dirname)

    if isdir:
        if must_not_exist:
            raise OSError('{0} must not exist but exists'.format(dirname))
        elif renew:
            logger.warning('Deleting directory %s', dirname)
            if not dry: shutil.rmtree(dirname)
        else:
            logger.warning('%s already exists, not recreating', dirname)
            return
    _create_directory_no_checks(dirname, dry=dry)

def copy_file(src, dst, dry=None):
    if dry is None: dry = qondor.DRYMODE
    logger.info('Copying %s --> %s', src, dst)
    if not dry: shutil.copy(src, dst)

class switchdir(object):
    """
    Context manager to temporarily change the working directory.

    :param newdir: Directory to change into
    :type newdir: str
    :param dry: Don't actually change directory if set to True
    :type dry: bool, optional
    """
    def __init__(self, newdir, dry=None):
        super(switchdir, self).__init__()
        self.newdir = newdir
        self._backdir = os.getcwd()
        self._no_need_to_change = (self.newdir == self._backdir)
        self.dry = qondor.DRYMODE if dry is None else dry

    def __enter__(self):
        if self._no_need_to_change:
            logger.info('Already in right directory, no need to change')
            return
        logger.info('chdir to {0}'.format(self.newdir))
        if not self.dry: os.chdir(self.newdir)

    def __exit__(self, type, value, traceback):
        if self._no_need_to_change:
            return
        logger.info('chdir back to {0}'.format(self._backdir))
        if not self.dry: os.chdir(self._backdir)

def run_command(cmd, env=None, dry=None, shell=False):
    logger.warning('Issuing command: {0}'.format(' '.join(cmd) if not is_string(cmd) else cmd))
    if dry is None: dry = qondor.DRYMODE
    if dry: return

    if shell and not is_string(cmd):
        cmd = ' '.join(cmd)

    process = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        env=env,
        universal_newlines=True,
        shell=shell
        )

    output = []
    for stdout_line in iter(process.stdout.readline, ""):
        subprocess_logger.info(stdout_line.rstrip('\n'))
        output.append(stdout_line)
    process.stdout.close()
    process.wait()
    returncode = process.returncode

    if returncode == 0:
        logger.info('Command exited with status 0 - all good')
    else:
        logger.error('Exit status {0} for command: {1}'.format(returncode, cmd))
        raise subprocess.CalledProcessError(cmd, returncode)
    return output

def run_multiple_commands(cmds, env=None, dry=None):
    logger.info('Sending cmds:\n{0}'.format(pprint.pformat(cmds)))
    if dry is None: dry = qondor.DRYMODE
    if dry:
        logger.info('Dry mode - not running command')
        return

    try:
        # Python 3 accepts encoding
        process = subprocess.Popen(
            'bash',
            stdin = subprocess.PIPE,
            stdout = subprocess.PIPE,
            stderr = subprocess.STDOUT,
            env = env,
            bufsize = 1,
            close_fds = True,
            encoding = 'utf8'
            )
    except TypeError:
        # Python 2 doesn't
        process = subprocess.Popen(
            'bash',
            stdin = subprocess.PIPE,
            stdout = subprocess.PIPE,
            stderr = subprocess.STDOUT,
            env = env,
            bufsize = 1,
            close_fds = True,
            universal_newlines = True
            )

    # Break on first error (stdin will still be written but execution will be stopped)
    process.stdin.write('set -e\n')
    process.stdin.flush()

    for cmd in cmds:
        if not(is_string(cmd)):
            cmd = ' '.join(cmd)
        if not(cmd.endswith('\n')):
            cmd += '\n'
        process.stdin.write(cmd)
        process.stdin.flush()
    process.stdin.close()

    output = []
    process.stdout.flush()
    for line in iter(process.stdout.readline, ""):
        if len(line) == 0: break
        line = line.rstrip('\n')
        subprocess_logger.info(line)
        output.append(line)

    process.stdout.close()
    process.wait()
    returncode = process.returncode

    if (returncode == 0):
        logger.info('Command exited with status 0 - all good')
        return output
    else:
        raise subprocess.CalledProcessError(cmd, returncode)


class DummyFile(object):
    def write(self, text):
        logger.debug('Writing: %s', text.replace('\n','\\n'))

@contextmanager
def openfile(*args, **kwargs):
    """
    Wrapper around the standard open(...) context, with the option of drymode
    """
    dry = kwargs.pop('dry', qondor.DRYMODE)
    try:
        if dry:
            yield DummyFile()
        else:
            with open(*args, **kwargs) as f:
                yield f
    finally:
        pass

def get_exitcode(cmd):
    if is_string(cmd): cmd = [cmd]
    logger.debug('Getting exit code for "%s"', ' '.join(cmd))
    FNULL = open(os.devnull, 'w')
    process = subprocess.Popen(cmd, stdout=FNULL, stderr=subprocess.STDOUT)
    process.communicate()[0]
    logger.debug('Got exit code %s', process.returncode)
    return process.returncode


def is_string(string):
    """
    Checks strictly whether `string` is a string
    Python 2 / 3 compatibility (https://stackoverflow.com/a/22679982/9209944)
    """
    try:
        basestring
    except NameError:
        basestring = str
    return isinstance(string, basestring)


def get_installation_path_of_module(module):
    logger.debug(
        'Trying to determine installation path of %s',
        module.__name__
        )
    try:
        logger.debug('Using pkg_resources')
        import pkg_resources
        distribution = pkg_resources.get_distribution(module.__name__)
        path = osp.abspath(distribution.location)
    except:
        logger.debug('From module.__path__')
        path = osp.abspath(module.__path__[0])
        if not osp.exists(path):
            logger.warning(
                'Path %s for module %s does not exist; reloading and trying again',
                path, module
                )
            reload(module)
            path = osp.abspath(module.__path__[0])
            if osp.exists(path):
                logger.warning(
                    'Path %s for module %s exists. '
                    'Did you chdir before calling this function? '
                    'module.__path__ is a relative path set at import time.',
                    path, module
                    )
            else:
                raise RuntimeError(
                    'Found no way of determining where %s is installed',
                    module.__name__
                    )
    logger.info(
        'Determined installation path of %s: %s',
        module.__name__, path
        )
    return path


def tarball_python_module(module, outdir=None, allow_uncommitted=True, dry=None, assume_pypi=True):
    """
    Takes a python module or a path to a file of said module, and attempts to make an installable
    pypi-style package tarball out of it.
    If assume_pypi is True, it will look for the pip installation directory of the package, and check
    whether there is a setup.py.
    Otherwise, it will look for the top-level git repository and make a tarball from that, including
    only files that are tracked by git. Uncommitted changes are included, unless allowed_uncommitted 
    is set to False.
    """
    if dry is None: dry = qondor.DRYMODE
    outdir = os.getcwd() if outdir is None else outdir
    outdir = osp.abspath(outdir)

    # Input variable may be a path
    if is_string(module):
        # Treat the input variable as a path
        path = module
    else:
        path = get_installation_path_of_module(module)

    # Make sure path exists and is a directory
    if not osp.exists(path):
        logger.error('Path %s does not seem to exist; cwd: %s', path, os.getcwd())
        raise OSError('{0} is not a valid path'.format(path))
    elif osp.isfile(path):
        path = osp.dirname(path)

    if assume_pypi:
        # Make sure it has a setup.py
        setuppy = osp.join(path, 'setup.py')
        if not osp.isfile(setuppy):
            raise OSError('Expected file {0} to exist; not a valid pypi package')
        outfile = osp.join(outdir, osp.basename(path) + '.tar')
        logger.info('Creating tarball from directory %s --> %s', path, outfile)
        if not dry:
            with switchdir(path):
                run_command([
                    'tar', '-cvf', outfile, '.',
                    '--exclude', '*/lib/python*',
                    '--exclude', '*/include/python*',
                    '--exclude', '*/bin/python*',
                    '--exclude', '*.egg-info*',
                    '--exclude', '*.pyc',
                    '--exclude', '*/.git'
                    ])
    else:
        logger.info('Package %s: Using top level git to create a tarball', path)
        # Get the top-level git dir
        with switchdir(path):
            toplevel_git_dir = run_command(['git', 'rev-parse', '--show-toplevel'])[0].strip()
        # Fix the output name of the tarball
        outfile = osp.join(outdir, osp.basename(toplevel_git_dir) + '.tar')
        with switchdir(toplevel_git_dir):
            if allow_uncommitted:
                logger.info('Creating tarball for %s including uncommitted changes', toplevel_git_dir)
                # Create the tarball with uncommitted changes in it
                if not dry:
                    run_command(
                        'git ls-files -z | xargs -0 tar -cvf {0}'
                        .format(outfile),
                        shell=True
                        )
            else:
                # Check if there are uncommitted changes
                try:
                    run_command(['git', 'diff-index', '--quiet', 'HEAD', '--'])
                except subprocess.CalledProcessError:
                    logger.error(
                        'Uncommitted changes detected; it is unlikely you want a tarball '
                        'with some changes not committed.'
                        )
                    raise
                # Create the actual tarball of the latest commit
                if not dry: run_command(['git', 'archive', '-o', outfile, 'HEAD'])
    logger.info('Created tarball {0}'.format(outfile))
    return outfile


def extract_tarball(tarball, outdir='.', dry=None):
    """
    Extracts a tarball to outdir
    """
    if dry is None: dry = qondor.DRYMODE
    tarball = osp.abspath(tarball)
    outdir = osp.abspath(outdir)
    logger.warning(
        'Extracting {0} ==> {1}'
        .format(tarball, outdir)
        )
    cmd = [
        'tar', '-xvf', tarball,
        '-C', outdir
        ]
    run_command(cmd, dry=dry)


def extract_tarball_cmssw(tarball, outdir='.', dry=None):
    """
    Extracts a tarball to outdir, and returns the extracted CMSSW dir
    """
    if dry is None: dry = qondor.DRYMODE
    extract_tarball(tarball, outdir, dry)
    # return the CMSSW directory
    if dry: return 'CMSSW_dry'
    return [ d for d in glob.glob(osp.join(outdir, 'CMSSW*')) if not d.endswith('.gz')][0]


def check_is_cmssw_path(path):
    """
    Checks whether the passed path contains a CMSSW distribution.
    """
    abs_path = osp.abspath(path)
    if not osp.basename(path).startswith('CMSSW'):
        raise ValueError(
            'Expected {0} to start with "CMSSW" (path: {1})'
            .format(osp.basename(path), abs_path)
            )
    if not osp.isdir(path):
        raise OSError(
            '{0} is not a directory (path: {1})'
            .format(path, abs_path)
            )
    if not osp.isdir(osp.join(path, 'src')):
        raise OSError(
            '{0} is not a directory (path: {1})'
            .format(osp.join(path, 'src'), abs_path)
            )

def get_clean_env():
    env = os.environ.copy()
    for var in [
        'ROOTSYS',
        'PATH',
        'LD_LIBRARY_PATH',
        'DYLD_LIBRARY_PATH',
        'SHLIB_PATH',
        'LIBPATH',
        'PYTHONPATH',
        'MANPATH',
        'CMAKE_PREFIX_PATH',
        'JUPYTER_PATH',
        # Added due to ROOT-env.sh
        'CPLUS_INCLUDE_PATH', 'CXX',
        'ZLIB_HOME',          'CURL_HOME',
        'DAVIX_HOME',         'GSL_HOME',
        'SETUPTOOLS_HOME',    'FONTCONFIG_HOME',
        'CAIRO_HOME',         'SQLITE_HOME',
        'PIXMAN_HOME',        'FREETYPE_HOME',
        'TBB_HOME',           'FC',
        'PKG_CONFIG_HOME',    'VC_HOME',
        'PNG_HOME',           'FFTW_HOME',
        'BOOST_HOME',         'VDT_HOME',
        'ROOT_HOME',          'ZEROMQ_HOME',
        'LIBXML2_HOME',       'PKG_CONFIG_PATH',
        'EXPAT_HOME',         'COMPILER_PATH',
        'BLAS_HOME',          'R_HOME',
        'XROOTD_HOME',        'MYSQL_HOME',
        'GFAL_HOME',          'CC',
        'C_INCLUDE_PATH',     'PYTHON_HOME',
        'PYTHONHOME',         'ORACLE_HOME',
        'GPERF_HOME',         'SRM_IFCE_HOME',
        'NUMPY_HOME',         'DCAP_HOME',
        ]:
        if var in env: del env[var]
    return env


def convert_to_utc(local_time):
    """
    Converts a local time to UTC.
    Implemented for only a few basic timezones.
    Needs to be extended with pytz if this function
    get seriously used.
    """
    # See: https://stackoverflow.com/a/10854983/9209944
    delta = datetime.timedelta(
        seconds = time.timezone if (time.localtime().tm_isdst == 0) else time.altzone
        )
    new_time = local_time + delta
    logger.debug('Offsetting %s by %s hours to get to UTC %s', local_time, delta, new_time)
    return new_time

def get_now_utc():
    return convert_to_utc(datetime.datetime.now())


def sleep_until(runtime_utc, allowed_lateness=300, is_not_utc=False):
    if is_not_utc:
        runtime_utc = convert_to_utc(runtime_utc)
    now_utc = get_now_utc()
    logger.info('Current time (UTC):       %s', now_utc)
    logger.info('Scheduled run time (UTC): %s', runtime_utc)

    delta = runtime_utc - now_utc
    delta_seconds = abs(delta.total_seconds())

    if delta < datetime.timedelta(seconds=0):
        # The job is too late; runtime_utc has already passed
        if delta_seconds < allowed_lateness:
            logger.info(
                'Job is late by %s seconds, which is within the allowed window'
                ' of max %s seconds late',
                delta_seconds, allowed_lateness
                )
            return 0
        else:
            logger.error(
                'Job is late by %s seconds, which is OUTSIDE the allowed window'
                ' of max %s seconds late - throwing exception',
                delta_seconds, allowed_lateness
                )
            raise RuntimeError
    else:
        logger.info(
            'Job is early by %s seconds, sleeping', delta_seconds
            )
        time.sleep(delta_seconds)
        return 0


def check_proxy():
    """
    Asserts that the user has a grid proxy that is valid for at least 168 more hours (1 week)
    """
    # cmd = 'voms-proxy-info -exists -valid 168:00' # Check if there is an existing proxy for a full week
    try:
        proxy_valid = subprocess.check_output(['grid-proxy-info', '-exists', '-valid', '168:00']) == 0
        logger.info('Found a valid proxy')
    except subprocess.CalledProcessError:
        logger.error(
            'Grid proxy is not valid for at least 1 week. Renew it using:\n'
            'voms-proxy-init -voms cms -valid 192:00'
            )
        raise

def dist_is_editable(dist):
    """
    Is distribution an editable install?
    see: https://stackoverflow.com/a/42583363/9209944
    """
    for path_item in sys.path:
        egg_link = osp.join(path_item, dist.project_name + '.egg-link')
        if osp.isfile(egg_link):
            return True
    return False


def _iter_chunkify_nrange(list_length, n_chunks):
    """
    Makes n_chunks chunks out of a range(list_length) list.
    Returns empty lists if n_chunks > list_length.
    """
    n_per_chunk_f = float(list_length) / n_chunks
    boundaries = [ (i*n_per_chunk_f, (i+1)*n_per_chunk_f) for i in range(n_chunks) ]
    for left, right in boundaries:
        indices_in_chunk = []
        for i in range(list_length):
            if i >= left and i < right:
                indices_in_chunk.append(i)
        yield indices_in_chunk

def iter_chunkify(mylist, n_chunks):
    """
    Makes n_chunks chunks out of mylist.
    """
    for indices in _iter_chunkify_nrange(len(mylist), n_chunks):
        yield [ mylist[i] for i in indices ]

def chunkify(mylist, n_chunks):
    return list(iter_chunkify(mylist, n_chunks))

def get_ith_chunk(mylist, n_chunks, i_chunk):
    return chunkify(mylist, n_chunks)[i_chunk]
